# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from contextlib import nullcontext
from unittest import mock
from urllib.parse import parse_qs, urlsplit

import pytest
from google.cloud.logging import Resource
from google.cloud.logging_v2.types import ListLogEntriesRequest, ListLogEntriesResponse, LogEntry

from airflow.providers.google.cloud.log.stackdriver_task_handler import StackdriverTaskHandler
from airflow.utils import timezone
from airflow.utils.state import TaskInstanceState

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.db import clear_db_dags, clear_db_runs
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS


def _create_list_log_entries_response_mock(messages, token):
    return ListLogEntriesResponse(
        entries=[LogEntry(json_payload={"message": message}) for message in messages], next_page_token=token
    )


@pytest.fixture
def clean_stackdriver_handlers():
    yield
    for handler_ref in reversed(logging._handlerList[:]):
        handler = handler_ref()
        if isinstance(handler, StackdriverTaskHandler):
            logging._removeHandlerRef(handler_ref)
            del handler


@pytest.mark.usefixtures("clean_stackdriver_handlers")
@mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
@mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.gcp_logging.Client")
def test_should_pass_message_to_client(mock_client, mock_get_creds_and_project_id):
    mock_get_creds_and_project_id.return_value = ("creds", "project_id")

    transport_type = mock.MagicMock()
    stackdriver_task_handler = StackdriverTaskHandler(transport=transport_type, labels={"key": "value"})
    logger = logging.getLogger("logger")
    logger.setLevel(logging.INFO)
    logger.addHandler(stackdriver_task_handler)

    logger.info("test-message")
    stackdriver_task_handler.flush()

    transport_type.assert_called_once_with(mock_client.return_value, "airflow")
    transport_type.return_value.send.assert_called_once_with(
        mock.ANY, "test-message", labels={"key": "value"}, resource=Resource(type="global", labels={})
    )
    mock_client.assert_called_once_with(credentials="creds", client_info=mock.ANY, project="project_id")


@pytest.mark.usefixtures("clean_stackdriver_handlers")
@mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
@mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.gcp_logging.Client")
def test_should_use_configured_log_name(mock_client, mock_get_creds_and_project_id):
    import importlib
    import logging

    from airflow import settings
    from airflow.config_templates import airflow_local_settings

    mock_get_creds_and_project_id.return_value = ("creds", "project_id")

    try:
        # this is needed for Airflow 2.8 and below where default settings are triggering warning on
        # extra "name" in the configuration of stackdriver handler. As of Airflow 2.9 this warning is not
        # emitted.
        context_manager = nullcontext()
        with context_manager:
            with conf_vars(
                {
                    ("logging", "remote_logging"): "True",
                    ("logging", "remote_base_log_folder"): "stackdriver://host/path",
                }
            ):
                importlib.reload(airflow_local_settings)
                settings.configure_logging()

                logger = logging.getLogger("airflow.task")
                handler = logger.handlers[0]
                assert isinstance(handler, StackdriverTaskHandler)
                with mock.patch.object(handler, "transport_type") as transport_type_mock:
                    logger.error("foo")
                    transport_type_mock.assert_called_once_with(mock_client.return_value, "path")
    finally:
        importlib.reload(airflow_local_settings)
        settings.configure_logging()


@pytest.mark.db_test
class TestStackdriverLoggingHandlerTask:
    DAG_ID = "dag_for_testing_stackdriver_file_task_handler"
    TASK_ID = "task_for_testing_stackdriver_task_handler"

    @pytest.fixture(autouse=True)
    def task_instance(self, create_task_instance, clean_stackdriver_handlers):
        self.ti = create_task_instance(
            dag_id=self.DAG_ID,
            task_id=self.TASK_ID,
            logical_date=timezone.datetime(2016, 1, 1),
            state=TaskInstanceState.RUNNING,
        )
        self.ti.try_number = 1
        self.ti.raw = False
        yield
        clear_db_runs()
        clear_db_dags()

    def _setup_handler(self, **handler_kwargs):
        self.transport_mock = mock.MagicMock()
        handler_kwargs = {"transport": self.transport_mock, **handler_kwargs}
        stackdriver_task_handler = StackdriverTaskHandler(**handler_kwargs)
        self.logger = logging.getLogger("logger")
        self.logger.addHandler(stackdriver_task_handler)
        return stackdriver_task_handler

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.gcp_logging.Client")
    def test_should_set_labels(self, mock_client, mock_get_creds_and_project_id):
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        stackdriver_task_handler = self._setup_handler()
        stackdriver_task_handler.set_context(self.ti)

        self.logger.info("test-message")
        stackdriver_task_handler.flush()

        date_key = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        labels = {
            "task_id": self.TASK_ID,
            "dag_id": self.DAG_ID,
            date_key: "2016-01-01T00:00:00+00:00",
            "try_number": "1",
        }
        resource = Resource(type="global", labels={})
        self.transport_mock.return_value.send.assert_called_once_with(
            mock.ANY, "test-message", labels=labels, resource=resource
        )

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.gcp_logging.Client")
    def test_should_append_labels(self, mock_client, mock_get_creds_and_project_id):
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        stackdriver_task_handler = self._setup_handler(
            labels={"product.googleapis.com/task_id": "test-value"},
        )
        stackdriver_task_handler.set_context(self.ti)

        self.logger.info("test-message")
        stackdriver_task_handler.flush()

        date_key = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        labels = {
            "task_id": self.TASK_ID,
            "dag_id": self.DAG_ID,
            date_key: "2016-01-01T00:00:00+00:00",
            "try_number": "1",
            "product.googleapis.com/task_id": "test-value",
        }
        resource = Resource(type="global", labels={})
        self.transport_mock.return_value.send.assert_called_once_with(
            mock.ANY, "test-message", labels=labels, resource=resource
        )

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_read_logs_for_all_try(self, mock_client, mock_get_creds_and_project_id):
        mock_client.return_value.list_log_entries.return_value.pages = iter(
            [_create_list_log_entries_response_mock(["MSG1", "MSG2"], None)]
        )
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        stackdriver_task_handler = self._setup_handler()
        logs, metadata = stackdriver_task_handler.read(self.ti)

        date_label = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"

        filter_str = (
            'resource.type="global"\n'
            'logName="projects/project_id/logs/airflow"\n'
            'labels.task_id="task_for_testing_stackdriver_task_handler"\n'
            'labels.dag_id="dag_for_testing_stackdriver_file_task_handler"\n'
            f'labels.{date_label}="2016-01-01T00:00:00+00:00"'
        )
        mock_client.return_value.list_log_entries.assert_called_once_with(
            request=ListLogEntriesRequest(
                resource_names=["projects/project_id"],
                filter=filter_str,
                order_by="timestamp asc",
                page_size=1000,
                page_token=None,
            )
        )
        assert logs == [(("default-hostname", "MSG1\nMSG2"),)]
        assert metadata == [{"end_of_log": True}]

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_read_logs_for_task_with_quote(self, mock_client, mock_get_creds_and_project_id):
        mock_client.return_value.list_log_entries.return_value.pages = iter(
            [_create_list_log_entries_response_mock(["MSG1", "MSG2"], None)]
        )
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        self.ti.task_id = 'K"OT'
        stackdriver_task_handler = self._setup_handler()

        logs, metadata = stackdriver_task_handler.read(self.ti)
        date_label = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        filter_str = (
            'resource.type="global"\n'
            'logName="projects/project_id/logs/airflow"\n'
            'labels.task_id="K\\"OT"\n'
            'labels.dag_id="dag_for_testing_stackdriver_file_task_handler"\n'
            f'labels.{date_label}="2016-01-01T00:00:00+00:00"'
        )
        mock_client.return_value.list_log_entries.assert_called_once_with(
            request=ListLogEntriesRequest(
                resource_names=["projects/project_id"],
                filter=filter_str,
                order_by="timestamp asc",
                page_size=1000,
                page_token=None,
            )
        )
        assert logs == [(("default-hostname", "MSG1\nMSG2"),)]
        assert metadata == [{"end_of_log": True}]

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_read_logs_for_single_try(self, mock_client, mock_get_creds_and_project_id):
        mock_client.return_value.list_log_entries.return_value.pages = iter(
            [_create_list_log_entries_response_mock(["MSG1", "MSG2"], None)]
        )
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")
        stackdriver_task_handler = self._setup_handler()

        logs, metadata = stackdriver_task_handler.read(self.ti, 3)
        date_label = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        filter_str = (
            'resource.type="global"\n'
            'logName="projects/project_id/logs/airflow"\n'
            'labels.task_id="task_for_testing_stackdriver_task_handler"\n'
            'labels.dag_id="dag_for_testing_stackdriver_file_task_handler"\n'
            f'labels.{date_label}="2016-01-01T00:00:00+00:00"\n'
            'labels.try_number="3"'
        )
        mock_client.return_value.list_log_entries.assert_called_once_with(
            request=ListLogEntriesRequest(
                resource_names=["projects/project_id"],
                filter=filter_str,
                order_by="timestamp asc",
                page_size=1000,
                page_token=None,
            )
        )
        assert logs == [(("default-hostname", "MSG1\nMSG2"),)]
        assert metadata == [{"end_of_log": True}]

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_read_logs_with_pagination(self, mock_client, mock_get_creds_and_project_id):
        mock_client.return_value.list_log_entries.side_effect = [
            mock.MagicMock(pages=iter([_create_list_log_entries_response_mock(["MSG1", "MSG2"], "TOKEN1")])),
            mock.MagicMock(pages=iter([_create_list_log_entries_response_mock(["MSG3", "MSG4"], None)])),
        ]
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")
        stackdriver_task_handler = self._setup_handler()

        logs, metadata1 = stackdriver_task_handler.read(self.ti, 3)
        date_label = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        filter_str = (
            'resource.type="global"\n'
            'logName="projects/project_id/logs/airflow"\n'
            'labels.task_id="task_for_testing_stackdriver_task_handler"\n'
            'labels.dag_id="dag_for_testing_stackdriver_file_task_handler"\n'
            f'labels.{date_label}="2016-01-01T00:00:00+00:00"\n'
            'labels.try_number="3"'
        )
        mock_client.return_value.list_log_entries.assert_called_once_with(
            request=ListLogEntriesRequest(
                resource_names=["projects/project_id"],
                filter=filter_str,
                order_by="timestamp asc",
                page_size=1000,
                page_token=None,
            )
        )
        assert logs == [(("default-hostname", "MSG1\nMSG2"),)]
        assert metadata1 == [{"end_of_log": False, "next_page_token": "TOKEN1"}]

        mock_client.return_value.list_log_entries.return_value.next_page_token = None
        logs, metadata2 = stackdriver_task_handler.read(self.ti, 3, metadata1[0])

        mock_client.return_value.list_log_entries.assert_called_with(
            request=ListLogEntriesRequest(
                resource_names=["projects/project_id"],
                filter=(
                    'resource.type="global"\n'
                    'logName="projects/project_id/logs/airflow"\n'
                    'labels.task_id="task_for_testing_stackdriver_task_handler"\n'
                    'labels.dag_id="dag_for_testing_stackdriver_file_task_handler"\n'
                    f'labels.{date_label}="2016-01-01T00:00:00+00:00"\n'
                    'labels.try_number="3"'
                ),
                order_by="timestamp asc",
                page_size=1000,
                page_token="TOKEN1",
            )
        )
        assert logs == [(("default-hostname", "MSG3\nMSG4"),)]
        assert metadata2 == [{"end_of_log": True}]

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_read_logs_with_download(self, mock_client, mock_get_creds_and_project_id):
        mock_client.return_value.list_log_entries.side_effect = [
            mock.MagicMock(pages=iter([_create_list_log_entries_response_mock(["MSG1", "MSG2"], "TOKEN1")])),
            mock.MagicMock(pages=iter([_create_list_log_entries_response_mock(["MSG3", "MSG4"], None)])),
        ]
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        stackdriver_task_handler = self._setup_handler()
        logs, metadata1 = stackdriver_task_handler.read(self.ti, 3, {"download_logs": True})

        assert logs == [(("default-hostname", "MSG1\nMSG2\nMSG3\nMSG4"),)]
        assert metadata1 == [{"end_of_log": True}]

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_read_logs_with_custom_resources(self, mock_client, mock_get_creds_and_project_id):
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")
        resource = Resource(
            type="cloud_composer_environment",
            labels={
                "environment.name": "test-instance",
                "location": "europe-west-3",
                "project_id": "project_id",
            },
        )
        stackdriver_task_handler = self._setup_handler(resource=resource)

        entry = mock.MagicMock(json_payload={"message": "TEXT"})
        page = mock.MagicMock(entries=[entry, entry], next_page_token=None)
        mock_client.return_value.list_log_entries.return_value.pages = iter([page])

        logs, metadata = stackdriver_task_handler.read(self.ti)
        date_label = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        filter_str = (
            'resource.type="cloud_composer_environment"\n'
            'logName="projects/project_id/logs/airflow"\n'
            'resource.labels."environment.name"="test-instance"\n'
            'resource.labels.location="europe-west-3"\n'
            'resource.labels.project_id="project_id"\n'
            'labels.task_id="task_for_testing_stackdriver_task_handler"\n'
            'labels.dag_id="dag_for_testing_stackdriver_file_task_handler"\n'
            f'labels.{date_label}="2016-01-01T00:00:00+00:00"'
        )
        mock_client.return_value.list_log_entries.assert_called_once_with(
            request=ListLogEntriesRequest(
                resource_names=["projects/project_id"],
                filter=filter_str,
                order_by="timestamp asc",
                page_size=1000,
                page_token=None,
            )
        )
        assert logs == [(("default-hostname", "TEXT\nTEXT"),)]
        assert metadata == [{"end_of_log": True}]

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.gcp_logging.Client")
    def test_should_use_credentials(self, mock_client, mock_get_creds_and_project_id):
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        stackdriver_task_handler = StackdriverTaskHandler(gcp_key_path="KEY_PATH")
        client = stackdriver_task_handler._client

        mock_get_creds_and_project_id.assert_called_once_with(
            disable_logging=True,
            key_path="KEY_PATH",
            scopes=frozenset(
                {
                    "https://www.googleapis.com/auth/logging.write",
                    "https://www.googleapis.com/auth/logging.read",
                }
            ),
        )
        mock_client.assert_called_once_with(credentials="creds", client_info=mock.ANY, project="project_id")
        assert mock_client.return_value == client

    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.get_credentials_and_project_id")
    @mock.patch("airflow.providers.google.cloud.log.stackdriver_task_handler.LoggingServiceV2Client")
    def test_should_return_valid_external_url(self, mock_client, mock_get_creds_and_project_id):
        mock_get_creds_and_project_id.return_value = ("creds", "project_id")

        stackdriver_task_handler = StackdriverTaskHandler(gcp_key_path="KEY_PATH")
        url = stackdriver_task_handler.get_external_log_url(self.ti, self.ti.try_number)

        parsed_url = urlsplit(url)
        parsed_qs = parse_qs(parsed_url.query)
        assert parsed_url.scheme == "https"
        assert parsed_url.netloc == "console.cloud.google.com"
        assert parsed_url.path == "/logs/viewer"
        assert {"project", "interval", "resource", "advancedFilter"} == set(parsed_qs.keys())
        assert "global" in parsed_qs["resource"]

        filter_params = parsed_qs["advancedFilter"][0].splitlines()
        date_label = "logical_date" if AIRFLOW_V_3_0_PLUS else "execution_date"
        expected_filter = [
            'resource.type="global"',
            'logName="projects/project_id/logs/airflow"',
            f'labels.task_id="{self.ti.task_id}"',
            f'labels.dag_id="{self.DAG_ID}"',
            f'labels.{date_label}="{self.ti.logical_date.isoformat() if AIRFLOW_V_3_0_PLUS else self.ti.execution_date.isoformat()}"',
            f'labels.try_number="{self.ti.try_number}"',
        ]
        assert set(expected_filter) == set(filter_params)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from contextlib import nullcontext as does_not_raise
from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.looker import JobStatus, LookerHook
from airflow.version import version

try:
    import importlib.util

    if not importlib.util.find_spec("airflow.sdk.bases.hook"):
        raise ImportError

    BASEHOOK_PATCH_PATH = "airflow.sdk.bases.hook.BaseHook"
except ImportError:
    BASEHOOK_PATCH_PATH = "airflow.hooks.base.BaseHook"
HOOK_PATH = "airflow.providers.google.cloud.hooks.looker.LookerHook.{}"

JOB_ID = "test-id"
TASK_ID = "test-task-id"
MODEL = "test_model"
VIEW = "test_view"
SOURCE = f"airflow:{version}"

CONN_EXTRA = {"verify_ssl": "true", "timeout": "120"}


class TestLookerHook:
    def setup_method(self):
        with mock.patch(f"{BASEHOOK_PATCH_PATH}.get_connection") as conn:
            conn.return_value.extra_dejson = CONN_EXTRA
            self.hook = LookerHook(looker_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.looker.requests_transport")
    def test_get_looker_sdk(self, _):
        """
        Test that get_looker_sdk is setting up the sdk properly

        Note: `requests_transport` is mocked so we don't have to test
        looker_sdk's functionality, just LookerHook's usage of it
        """
        self.hook.get_connection = mock.MagicMock()
        sdk = self.hook.get_looker_sdk()

        # Attempting to use the instantiated SDK should not throw an error
        with does_not_raise():
            sdk.get(path="/", structure=None)
            sdk.delete(path="/", structure=None)

            # The post/patch/put methods call a method internally to serialize
            # the body if LookerHook sets the wrong serialize function on init
            # there will be TypeErrors thrown
            # The body we pass here must be a list, dict, or model.Model
            sdk.post(path="/", structure=None, body=[])
            sdk.patch(path="/", structure=None, body=[])
            sdk.put(path="/", structure=None, body=[])

    @mock.patch(HOOK_PATH.format("pdt_build_status"))
    def test_wait_for_job(self, mock_pdt_build_status):
        # replace pdt_build_status invocation with mock status
        mock_pdt_build_status.side_effect = [
            {"status": JobStatus.RUNNING.value},
            {"status": JobStatus.ERROR.value, "message": "test"},
        ]

        # call hook in mock context (w/ no wait b/w job checks)
        with pytest.raises(AirflowException):
            self.hook.wait_for_job(
                materialization_id=JOB_ID,
                wait_time=0,
            )

        # assert pdt_build_status called twice: first RUNNING, then ERROR
        calls = [
            mock.call(materialization_id=JOB_ID),
            mock.call(materialization_id=JOB_ID),
        ]
        mock_pdt_build_status.assert_has_calls(calls)

    @mock.patch(HOOK_PATH.format("get_looker_sdk"))
    def test_check_pdt_build(self, mock_sdk):
        # call hook in mock context
        self.hook.check_pdt_build(materialization_id=JOB_ID)

        # assert sdk constructor called once
        mock_sdk.assert_called_once_with()

        # assert sdk.check_pdt_build called once
        mock_sdk.return_value.check_pdt_build.assert_called_once_with(materialization_id=JOB_ID)

    @mock.patch(HOOK_PATH.format("get_looker_sdk"))
    def test_start_pdt_build(self, mock_sdk):
        # replace looker version invocation with mock response
        mock_sdk.return_value.versions.return_value.looker_release_version = "22.2.0"

        # call hook in mock context
        self.hook.start_pdt_build(
            model=MODEL,
            view=VIEW,
        )

        # assert sdk constructor called once
        mock_sdk.assert_called_once_with()

        # assert sdk.start_pdt_build called once
        mock_sdk.return_value.start_pdt_build.assert_called_once_with(
            model_name=MODEL,
            view_name=VIEW,
            source=SOURCE,
        )

    @mock.patch(HOOK_PATH.format("get_looker_sdk"))
    def test_stop_pdt_build(self, mock_sdk):
        # call hook in mock context
        self.hook.stop_pdt_build(materialization_id=JOB_ID)

        # assert sdk constructor called once
        mock_sdk.assert_called_once_with()

        # assert sdk.stop_pdt_build called once
        mock_sdk.return_value.stop_pdt_build.assert_called_once_with(
            materialization_id=JOB_ID,
            source=SOURCE,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import base64
import json
import os
import platform
import tempfile
from unittest import mock
from unittest.mock import PropertyMock, call, mock_open
from urllib.parse import parse_qsl, unquote, urlsplit

import aiohttp
import httplib2
import pytest
from aiohttp.helpers import TimerNoop
from googleapiclient.errors import HttpError
from yarl import URL

from airflow.exceptions import AirflowException

from tests_common.test_utils.version_compat import AIRFLOW_V_3_1_PLUS

if AIRFLOW_V_3_1_PLUS:
    from airflow.sdk import Connection
else:
    from airflow.models import Connection  # type: ignore[assignment,attr-defined,no-redef]
from airflow.providers.google.cloud.hooks.cloud_sql import (
    CloudSQLAsyncHook,
    CloudSQLDatabaseHook,
    CloudSQLHook,
    CloudSqlProxyRunner,
)

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

HOOK_STR = "airflow.providers.google.cloud.hooks.cloud_sql.{}"
PROJECT_ID = "test_project_id"
OPERATION_NAME = "test_operation_name"
OPERATION_URL = (
    f"https://sqladmin.googleapis.com/sql/v1beta4/projects/{PROJECT_ID}/operations/{OPERATION_NAME}"
)
SSL_CERT = "sslcert.pem"
SSL_KEY = "sslkey.pem"
SSL_ROOT_CERT = "sslrootcert.pem"
CONNECTION_ID = "test-conn-id"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
SECRET_ID = "test-secret-id"


@pytest.fixture
def hook_async():
    with mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    ):
        yield CloudSQLAsyncHook()


def session():
    return mock.Mock()


class TestGcpSqlHookDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.cloudsql_hook = CloudSQLHook(api_version="v1", gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    def test_instance_import_exception(self, mock_get_credentials):
        self.cloudsql_hook.get_conn = mock.Mock(
            side_effect=HttpError(resp=httplib2.Response({"status": 400}), content=b"Error content")
        )
        with pytest.raises(AirflowException) as ctx:
            self.cloudsql_hook.import_instance(instance="instance", body={})
        err = ctx.value
        assert "Importing instance " in str(err)
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    def test_instance_export_exception(self, mock_get_credentials):
        self.cloudsql_hook.get_conn = mock.Mock(
            side_effect=HttpError(resp=httplib2.Response({"status": 400}), content=b"Error content")
        )
        with pytest.raises(HttpError) as ctx:
            self.cloudsql_hook.export_instance(instance="instance", body={})
        err = ctx.value
        assert err.resp.status == 400
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_instance_import(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        import_method = get_conn.return_value.instances.return_value.import_
        execute_method = import_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.import_instance(instance="instance", body={})

        import_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_instance_export(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        export_method = get_conn.return_value.instances.return_value.export
        execute_method = export_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.export_instance(instance="instance", body={})

        export_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        assert mock_get_credentials.call_count == 1

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_instance_export_with_in_progress_retry(self, wait_for_operation_to_complete, get_conn):
        export_method = get_conn.return_value.instances.return_value.export
        execute_method = export_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        with pytest.raises(HttpError):
            self.cloudsql_hook.export_instance(project_id="example-project", instance="instance", body={})
        wait_for_operation_to_complete.assert_not_called()

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_get_instance(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        get_method = get_conn.return_value.instances.return_value.get
        execute_method = get_method.return_value.execute
        execute_method.return_value = {"name": "instance"}
        wait_for_operation_to_complete.return_value = None
        res = self.cloudsql_hook.get_instance(instance="instance")
        assert res is not None
        assert res["name"] == "instance"
        get_method.assert_called_once_with(instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_not_called()
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_create_instance(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        insert_method = get_conn.return_value.instances.return_value.insert
        execute_method = insert_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.create_instance(body={})

        insert_method.assert_called_once_with(body={}, project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_create_instance_with_in_progress_retry(
        self, wait_for_operation_to_complete, get_conn, mock_get_credentials
    ):
        insert_method = get_conn.return_value.instances.return_value.insert
        execute_method = insert_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.create_instance(body={})

        assert mock_get_credentials.call_count == 1
        assert insert_method.call_count == 2
        assert execute_method.call_count == 2
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_patch_instance_with_in_progress_retry(
        self, wait_for_operation_to_complete, get_conn, mock_get_credentials
    ):
        patch_method = get_conn.return_value.instances.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.patch_instance(instance="instance", body={})

        assert mock_get_credentials.call_count == 1
        assert patch_method.call_count == 2
        assert execute_method.call_count == 2
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_patch_instance(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        patch_method = get_conn.return_value.instances.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.patch_instance(instance="instance", body={})

        patch_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_delete_instance(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        delete_method = get_conn.return_value.instances.return_value.delete
        execute_method = delete_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.delete_instance(instance="instance")

        delete_method.assert_called_once_with(instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project", time_to_sleep=5
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_delete_instance_with_in_progress_retry(
        self, wait_for_operation_to_complete, get_conn, mock_get_credentials
    ):
        delete_method = get_conn.return_value.instances.return_value.delete
        execute_method = delete_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.delete_instance(instance="instance")

        assert mock_get_credentials.call_count == 1
        assert delete_method.call_count == 2
        assert execute_method.call_count == 2
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project", time_to_sleep=5
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_instance_clone(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        clone_method = get_conn.return_value.instances.return_value.clone
        execute_method = clone_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        body = {
            "cloneContext": {
                "kind": "sql#cloneContext",
                "destinationInstanceName": "clonedInstance",
            }
        }
        self.cloudsql_hook.clone_instance(instance="instance", body=body)

        clone_method.assert_called_once_with(instance="instance", project="example-project", body=body)
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_get_database(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        get_method = get_conn.return_value.databases.return_value.get
        execute_method = get_method.return_value.execute
        execute_method.return_value = {"name": "database"}
        wait_for_operation_to_complete.return_value = None
        res = self.cloudsql_hook.get_database(database="database", instance="instance")
        assert res is not None
        assert res["name"] == "database"
        get_method.assert_called_once_with(
            instance="instance", database="database", project="example-project"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_not_called()
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_create_database(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        insert_method = get_conn.return_value.databases.return_value.insert
        execute_method = insert_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.create_database(instance="instance", body={})

        insert_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_create_database_with_in_progress_retry(
        self, wait_for_operation_to_complete, get_conn, mock_get_credentials
    ):
        insert_method = get_conn.return_value.databases.return_value.insert
        execute_method = insert_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.create_database(instance="instance", body={})

        assert mock_get_credentials.call_count == 1
        assert insert_method.call_count == 2
        assert execute_method.call_count == 2
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_patch_database(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        patch_method = get_conn.return_value.databases.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.patch_database(instance="instance", database="database", body={})

        patch_method.assert_called_once_with(
            body={}, database="database", instance="instance", project="example-project"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_patch_database_with_in_progress_retry(
        self, wait_for_operation_to_complete, get_conn, mock_get_credentials
    ):
        patch_method = get_conn.return_value.databases.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.patch_database(instance="instance", database="database", body={})

        assert mock_get_credentials.call_count == 1
        assert patch_method.call_count == 2
        assert execute_method.call_count == 2
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_delete_database(self, wait_for_operation_to_complete, get_conn, mock_get_credentials):
        delete_method = get_conn.return_value.databases.return_value.delete
        execute_method = delete_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.delete_database(instance="instance", database="database")

        delete_method.assert_called_once_with(
            database="database", instance="instance", project="example-project"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )
        assert mock_get_credentials.call_count == 1

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_credentials_and_project_id",
        return_value=(mock.MagicMock(), "example-project"),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_delete_database_with_in_progress_retry(
        self, wait_for_operation_to_complete, get_conn, mock_get_credentials
    ):
        delete_method = get_conn.return_value.databases.return_value.delete
        execute_method = delete_method.return_value.execute
        execute_method.side_effect = [
            HttpError(
                resp=httplib2.Response({"status": 429}),
                content=b"Internal Server Error",
            ),
            {"name": "operation_id"},
        ]
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook.delete_database(instance="instance", database="database")

        assert mock_get_credentials.call_count == 1
        assert delete_method.call_count == 2
        assert execute_method.call_count == 2
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )


class TestGcpSqlHookNoDefaultProjectID:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.cloudsql_hook_no_default_project_id = CloudSQLHook(api_version="v1", gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_instance_import_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        import_method = get_conn.return_value.instances.return_value.import_
        execute_method = import_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.import_instance(
            project_id="example-project", instance="instance", body={}
        )
        import_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_instance_export_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        export_method = get_conn.return_value.instances.return_value.export
        execute_method = export_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.export_instance(
            project_id="example-project", instance="instance", body={}
        )
        export_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_get_instance_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        get_method = get_conn.return_value.instances.return_value.get
        execute_method = get_method.return_value.execute
        execute_method.return_value = {"name": "instance"}
        wait_for_operation_to_complete.return_value = None
        res = self.cloudsql_hook_no_default_project_id.get_instance(
            project_id="example-project", instance="instance"
        )
        assert res is not None
        assert res["name"] == "instance"
        get_method.assert_called_once_with(instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_not_called()

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_create_instance_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        insert_method = get_conn.return_value.instances.return_value.insert
        execute_method = insert_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.create_instance(project_id="example-project", body={})
        insert_method.assert_called_once_with(body={}, project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_patch_instance_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        patch_method = get_conn.return_value.instances.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.patch_instance(
            project_id="example-project", instance="instance", body={}
        )
        patch_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_delete_instance_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        delete_method = get_conn.return_value.instances.return_value.delete
        execute_method = delete_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.delete_instance(
            project_id="example-project", instance="instance"
        )
        delete_method.assert_called_once_with(instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project", time_to_sleep=5
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_get_database_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        get_method = get_conn.return_value.databases.return_value.get
        execute_method = get_method.return_value.execute
        execute_method.return_value = {"name": "database"}
        wait_for_operation_to_complete.return_value = None
        res = self.cloudsql_hook_no_default_project_id.get_database(
            project_id="example-project", database="database", instance="instance"
        )
        assert res is not None
        assert res["name"] == "database"
        get_method.assert_called_once_with(
            instance="instance", database="database", project="example-project"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_not_called()

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_create_database_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        insert_method = get_conn.return_value.databases.return_value.insert
        execute_method = insert_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.create_database(
            project_id="example-project", instance="instance", body={}
        )
        insert_method.assert_called_once_with(body={}, instance="instance", project="example-project")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_patch_database_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        patch_method = get_conn.return_value.databases.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.patch_database(
            project_id="example-project", instance="instance", database="database", body={}
        )
        patch_method.assert_called_once_with(
            body={}, database="database", instance="instance", project="example-project"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook._wait_for_operation_to_complete")
    def test_delete_database_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn, mock_project_id
    ):
        delete_method = get_conn.return_value.databases.return_value.delete
        execute_method = delete_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        self.cloudsql_hook_no_default_project_id.delete_database(
            project_id="example-project", instance="instance", database="database"
        )
        delete_method.assert_called_once_with(
            database="database", instance="instance", project="example-project"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project"
        )


def _parse_from_uri(uri: str):
    connection_parameters = {}
    uri_parts = urlsplit(uri)
    connection_parameters["conn_type"] = uri_parts.scheme
    rest_of_the_url = uri.replace(f"{uri_parts.scheme}://", "//")
    uri_parts = urlsplit(rest_of_the_url)
    host = unquote(uri_parts.hostname or "")
    connection_parameters["host"] = host
    quoted_schema = uri_parts.path[1:]
    connection_parameters["schema"] = unquote(quoted_schema) if quoted_schema else ""
    connection_parameters["login"] = unquote(uri_parts.username) if uri_parts.username else ""
    connection_parameters["password"] = unquote(uri_parts.password) if uri_parts.password else ""
    connection_parameters["port"] = uri_parts.port  # type: ignore[assignment]
    if uri_parts.query:
        query = dict(parse_qsl(uri_parts.query, keep_blank_values=True))
        connection_parameters["extra"] = json.dumps(query)
    return connection_parameters


class TestCloudSqlDatabaseHook:
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_validate_ssl_certs_no_ssl(self, get_connection):
        connection = Connection(
            conn_id="my_test_connection",
            conn_type="gcpcloudsqldb",
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = json.dumps(
                {"location": "test", "instance": "instance", "database_type": "postgres"}
            )
        else:
            connection.set_extra(
                json.dumps({"location": "test", "instance": "instance", "database_type": "postgres"})
            )
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        hook.validate_ssl_certs()

    @pytest.mark.parametrize(
        "cert_dict",
        [
            {},
            {"sslcert": "cert_file.pem"},
            {"sslkey": "cert_key.pem"},
            {"sslrootcert": "root_cert_file.pem"},
            {"sslcert": "cert_file.pem", "sslkey": "cert_key.pem"},
            {"sslrootcert": "root_cert_file.pem", "sslkey": "cert_key.pem"},
            {"sslrootcert": "root_cert_file.pem", "sslcert": "cert_file.pem"},
        ],
    )
    @mock.patch("os.path.isfile")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook._set_temporary_ssl_file")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_validate_ssl_certs_missing_cert_params(
        self, get_connection, mock_set_temporary_ssl_file, mock_is_file, cert_dict
    ):
        mock_is_file.side_effects = True
        mock_set_temporary_ssl_file.side_effect = cert_dict.values()
        connection = Connection(
            conn_id="my_test_connection",
            conn_type="gcpcloudsqldb",
        )
        extras = {"location": "test", "instance": "instance", "database_type": "postgres", "use_ssl": "True"}
        extras.update(cert_dict)
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = json.dumps(extras)
        else:
            connection.set_extra(json.dumps(extras))

        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        with pytest.raises(AirflowException) as ctx:
            hook.validate_ssl_certs()
        err = ctx.value
        assert "SSL connections requires" in str(err)

    @mock.patch("os.path.isfile")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook._set_temporary_ssl_file")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_validate_ssl_certs_with_ssl(
        self, get_connection, mock_set_temporary_ssl_file, mock_is_file
    ):
        connection = Connection(
            conn_id="my_test_connection",
            conn_type="gcpcloudsqldb",
        )
        mock_is_file.return_value = True
        mock_set_temporary_ssl_file.side_effect = [
            "/tmp/cert_file.pem",
            "/tmp/rootcert_file.pem",
            "/tmp/key_file.pem",
        ]
        extras = json.dumps(
            {
                "location": "test",
                "instance": "instance",
                "database_type": "postgres",
                "use_ssl": "True",
                "sslcert": "cert_file.pem",
                "sslrootcert": "rootcert_file.pem",
                "sslkey": "key_file.pem",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        hook.validate_ssl_certs()

    @mock.patch("os.path.isfile")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook._set_temporary_ssl_file")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_validate_ssl_certs_with_ssl_files_not_readable(
        self, get_connection, mock_set_temporary_ssl_file, mock_is_file
    ):
        connection = Connection(
            conn_id="my_test_connection",
            conn_type="gcpcloudsqldb",
        )
        mock_is_file.return_value = False
        mock_set_temporary_ssl_file.side_effect = [
            "/tmp/cert_file.pem",
            "/tmp/rootcert_file.pem",
            "/tmp/key_file.pem",
        ]
        extras = json.dumps(
            {
                "location": "test",
                "instance": "instance",
                "database_type": "postgres",
                "use_ssl": "True",
                "sslcert": "cert_file.pem",
                "sslrootcert": "rootcert_file.pem",
                "sslkey": "key_file.pem",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        with pytest.raises(AirflowException) as ctx:
            hook.validate_ssl_certs()
        err = ctx.value
        assert "must be a readable file" in str(err)

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.gettempdir")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_validate_socket_path_length_too_long(
        self, get_connection, gettempdir_mock
    ):
        gettempdir_mock.return_value = "/tmp"
        connection = Connection(
            conn_id="my_test_connection",
            conn_type="gcpcloudsqldb",
        )
        extras = json.dumps(
            {
                "location": "test",
                "instance": "very_long_instance_name_that_will_be_too_long_to_build_socket_length",
                "database_type": "postgres",
                "use_proxy": "True",
                "use_tcp": "False",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        with pytest.raises(AirflowException) as ctx:
            hook.validate_socket_path_length()
        err = ctx.value
        assert "The UNIX socket path length cannot exceed" in str(err)

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.gettempdir")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_validate_socket_path_length_not_too_long(
        self, get_connection, gettempdir_mock
    ):
        gettempdir_mock.return_value = "/tmp"
        connection = Connection(
            conn_id="my_test_connection",
            conn_type="gcpcloudsqldb",
        )
        extras = json.dumps(
            {
                "location": "test",
                "instance": "short_instance_name",
                "database_type": "postgres",
                "use_proxy": "True",
                "use_tcp": "False",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        hook.validate_socket_path_length()

    @pytest.mark.parametrize(
        "uri",
        [
            "http://:password@host:80/database",
            "http://user:@host:80/database",
            "http://user:password@/database",
            "http://user:password@host:80/",
            "http://user:password@/",
            "http://host:80/database",
            "http://host:80/",
        ],
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_create_connection_missing_fields(self, get_connection, uri):
        if AIRFLOW_V_3_1_PLUS:
            connection = Connection(conn_id="test_conn_id", **_parse_from_uri(uri))
        else:
            connection = Connection(uri=uri)
        params = {
            "location": "test",
            "instance": "instance",
            "database_type": "postgres",
            "use_proxy": "True",
            "use_tcp": "False",
        }
        extras = json.dumps(params)
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        with pytest.raises(AirflowException) as ctx:
            hook.create_connection()
        err = ctx.value
        assert "needs to be set in connection" in str(err)

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_get_sqlproxy_runner_no_proxy(self, get_connection):
        if AIRFLOW_V_3_1_PLUS:
            connection = Connection(
                conn_id="test_conn_id", **_parse_from_uri("http://user:password@host:80/database")
            )
        else:
            connection = Connection(uri="http://user:password@host:80/database")
        extras = json.dumps(
            {
                "location": "test",
                "instance": "instance",
                "database_type": "postgres",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        with pytest.raises(
            ValueError, match="Proxy runner can only be retrieved in case of use_proxy = True"
        ):
            hook.get_sqlproxy_runner()

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_get_sqlproxy_runner(self, get_connection):
        if AIRFLOW_V_3_1_PLUS:
            connection = Connection(
                conn_id="test_conn_id", **_parse_from_uri("http://user:password@host:80/database")
            )
        else:
            connection = Connection(uri="http://user:password@host:80/database")
        extras = json.dumps(
            {
                "location": "test",
                "instance": "instance",
                "database_type": "postgres",
                "use_proxy": "True",
                "use_tcp": "False",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        hook.create_connection()
        proxy_runner = hook.get_sqlproxy_runner()
        assert proxy_runner is not None

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_cloudsql_database_hook_get_database_hook(self, get_connection):
        if AIRFLOW_V_3_1_PLUS:
            connection = Connection(
                conn_id="test_conn_id", **_parse_from_uri("http://user:password@host:80/database")
            )
        else:
            connection = Connection(uri="http://user:password@host:80/database")
        extras = json.dumps(
            {
                "location": "test",
                "instance": "instance",
                "database_type": "postgres",
            }
        )
        if AIRFLOW_V_3_1_PLUS:
            connection.extra = extras
        else:
            connection.set_extra(extras)
        get_connection.return_value = connection
        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection", default_gcp_project_id="google_connection"
        )
        connection = hook.create_connection()
        db_hook = hook.get_database_hook(connection=connection)
        assert db_hook is not None

    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook._get_ssl_temporary_file_path"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_ssl_cert_properties(self, mock_get_connection, mock_get_ssl_temporary_file_path):
        def side_effect_func(cert_name, cert_path):
            return f"/tmp/certs/{cert_name}"

        mock_get_ssl_temporary_file_path.side_effect = side_effect_func
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson=dict(database_type="postgres", location="test", instance="instance")
        )

        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection",
            default_gcp_project_id="google_connection",
            ssl_cert=SSL_CERT,
            ssl_key=SSL_KEY,
            ssl_root_cert=SSL_ROOT_CERT,
        )
        sslcert = hook.sslcert
        sslkey = hook.sslkey
        sslrootcert = hook.sslrootcert

        assert hook.ssl_cert == SSL_CERT
        assert hook.ssl_key == SSL_KEY
        assert hook.ssl_root_cert == SSL_ROOT_CERT
        assert sslcert == "/tmp/certs/sslcert"
        assert sslkey == "/tmp/certs/sslkey"
        assert sslrootcert == "/tmp/certs/sslrootcert"
        mock_get_ssl_temporary_file_path.assert_has_calls(
            [
                call(cert_name="sslcert", cert_path=SSL_CERT),
                call(cert_name="sslkey", cert_path=SSL_KEY),
                call(cert_name="sslrootcert", cert_path=SSL_ROOT_CERT),
            ]
        )

    @pytest.mark.parametrize("ssl_name", ["sslcert", "sslkey", "sslrootcert"])
    @pytest.mark.parametrize(
        ("cert_value", "cert_path", "extra_cert_path"),
        [
            (None, None, "/connection/path/to/cert.pem"),
            (None, "/path/to/cert.pem", None),
            (None, "/path/to/cert.pem", "/connection/path/to/cert.pem"),
            (mock.MagicMock(), None, None),
            (mock.MagicMock(), None, "/connection/path/to/cert.pem"),
            (mock.MagicMock(), "/path/to/cert.pem", None),
            (mock.MagicMock(), "/path/to/cert.pem", "/connection/path/to/cert.pem"),
        ],
    )
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook._set_temporary_ssl_file"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook._get_cert_from_secret"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_get_ssl_temporary_file_path(
        self,
        mock_get_connection,
        mock_get_cert_from_secret,
        mock_set_temporary_ssl_file,
        cert_value,
        cert_path,
        extra_cert_path,
        ssl_name,
    ):
        expected_cert_file_path = cert_path
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
                ssl_name: extra_cert_path,
            }
        )
        mock_get_cert_from_secret.return_value = cert_value
        mock_set_temporary_ssl_file.return_value = expected_cert_file_path

        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection",
            default_gcp_project_id="google_connection",
            ssl_cert=SSL_CERT,
            ssl_key=SSL_KEY,
            ssl_root_cert=SSL_ROOT_CERT,
        )
        actual_cert_file_path = hook._get_ssl_temporary_file_path(cert_name=ssl_name, cert_path=cert_path)

        assert actual_cert_file_path == expected_cert_file_path
        assert hook.extras.get(ssl_name) == extra_cert_path
        mock_get_cert_from_secret.assert_called_once_with(ssl_name)
        mock_set_temporary_ssl_file.assert_called_once_with(
            cert_name=ssl_name, cert_path=cert_path or extra_cert_path, cert_value=cert_value
        )

    @pytest.mark.parametrize("ssl_name", ["sslcert", "sslkey", "sslrootcert"])
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook._set_temporary_ssl_file"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook._get_cert_from_secret"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_get_ssl_temporary_file_path_none(
        self,
        mock_get_connection,
        mock_get_cert_from_secret,
        mock_set_temporary_ssl_file,
        ssl_name,
    ):
        expected_cert_file_path = None
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        mock_get_cert_from_secret.return_value = None
        mock_set_temporary_ssl_file.return_value = expected_cert_file_path

        hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="cloudsql_connection",
            default_gcp_project_id="google_connection",
            ssl_cert=SSL_CERT,
            ssl_key=SSL_KEY,
            ssl_root_cert=SSL_ROOT_CERT,
        )
        actual_cert_file_path = hook._get_ssl_temporary_file_path(cert_name=ssl_name, cert_path=None)

        assert actual_cert_file_path == expected_cert_file_path
        assert hook.extras.get(ssl_name) is None
        mock_get_cert_from_secret.assert_called_once_with(ssl_name)
        assert not mock_set_temporary_ssl_file.called

    @pytest.mark.parametrize(
        ("cert_name", "cert_value"),
        [
            ["sslcert", SSL_CERT],
            ["sslkey", SSL_KEY],
            ["sslrootcert", SSL_ROOT_CERT],
        ],
    )
    @mock.patch(HOOK_STR.format("GoogleCloudSecretManagerHook"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_get_cert_from_secret(
        self,
        mock_get_connection,
        mock_secret_hook,
        cert_name,
        cert_value,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        mock_secret = mock_secret_hook.return_value.access_secret.return_value
        mock_secret.payload.data = base64.b64encode(json.dumps({cert_name: cert_value}).encode("ascii"))

        hook = CloudSQLDatabaseHook(
            gcp_conn_id=CONNECTION_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            default_gcp_project_id=PROJECT_ID,
            ssl_secret_id=SECRET_ID,
        )
        actual_cert_value = hook._get_cert_from_secret(cert_name=cert_name)

        assert actual_cert_value == cert_value
        mock_secret_hook.assert_called_once_with(
            gcp_conn_id=CONNECTION_ID, impersonation_chain=IMPERSONATION_CHAIN
        )
        mock_secret_hook.return_value.access_secret.assert_called_once_with(
            project_id=PROJECT_ID, secret_id=SECRET_ID
        )

    @pytest.mark.parametrize(
        ("cert_name", "cert_value"),
        [
            ["sslcert", SSL_CERT],
            ["sslkey", SSL_KEY],
            ["sslrootcert", SSL_ROOT_CERT],
        ],
    )
    @mock.patch(HOOK_STR.format("GoogleCloudSecretManagerHook"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_get_cert_from_secret_exception(
        self,
        mock_get_connection,
        mock_secret_hook,
        cert_name,
        cert_value,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        mock_secret = mock_secret_hook.return_value.access_secret.return_value
        mock_secret.payload.data = base64.b64encode(json.dumps({"wrong_key": cert_value}).encode("ascii"))

        hook = CloudSQLDatabaseHook(
            gcp_conn_id=CONNECTION_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            default_gcp_project_id=PROJECT_ID,
            ssl_secret_id=SECRET_ID,
        )

        with pytest.raises(AirflowException):
            hook._get_cert_from_secret(cert_name=cert_name)

        mock_secret_hook.assert_called_once_with(
            gcp_conn_id=CONNECTION_ID, impersonation_chain=IMPERSONATION_CHAIN
        )
        mock_secret_hook.return_value.access_secret.assert_called_once_with(
            project_id=PROJECT_ID, secret_id=SECRET_ID
        )

    @pytest.mark.parametrize("cert_name", ["sslcert", "sslkey", "sslrootcert"])
    @mock.patch(HOOK_STR.format("GoogleCloudSecretManagerHook"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_get_cert_from_secret_none(
        self,
        mock_get_connection,
        mock_secret_hook,
        cert_name,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )

        hook = CloudSQLDatabaseHook(
            gcp_conn_id=CONNECTION_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            default_gcp_project_id=PROJECT_ID,
        )
        actual_cert_value = hook._get_cert_from_secret(cert_name=cert_name)

        assert actual_cert_value is None
        assert not mock_secret_hook.called
        assert not mock_secret_hook.return_value.access_secret.called

    @pytest.mark.parametrize("cert_name", ["sslcert", "sslkey", "sslrootcert"])
    @mock.patch("builtins.open", new_callable=mock_open, read_data="test-data")
    @mock.patch(HOOK_STR.format("NamedTemporaryFile"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_set_temporary_ssl_file_cert_path(
        self,
        mock_get_connection,
        mock_named_temporary_file,
        mock_open_file,
        cert_name,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        expected_file_name = "/test/path/to/file"
        mock_named_temporary_file.return_value.name = expected_file_name
        source_cert_path = "/source/cert/path/to/file"

        hook = CloudSQLDatabaseHook()
        actual_path = hook._set_temporary_ssl_file(cert_name=cert_name, cert_path=source_cert_path)

        assert actual_path == expected_file_name
        mock_named_temporary_file.assert_called_once_with(mode="w+b", prefix="/tmp/certs/")
        mock_open_file.assert_has_calls([call(source_cert_path, "rb")])
        mock_named_temporary_file.return_value.write.assert_called_once_with("test-data")
        mock_named_temporary_file.return_value.flush.assert_called_once()

    @pytest.mark.parametrize("cert_name", ["sslcert", "sslkey", "sslrootcert"])
    @mock.patch(HOOK_STR.format("NamedTemporaryFile"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_set_temporary_ssl_file_cert_value(
        self,
        mock_get_connection,
        mock_named_temporary_file,
        cert_name,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        expected_file_name = "/test/path/to/file"
        mock_named_temporary_file.return_value.name = expected_file_name
        cert_value = "test-cert-value"

        hook = CloudSQLDatabaseHook()
        actual_path = hook._set_temporary_ssl_file(cert_name=cert_name, cert_value=cert_value)

        assert actual_path == expected_file_name
        mock_named_temporary_file.assert_called_once_with(mode="w+b", prefix="/tmp/certs/")
        mock_named_temporary_file.return_value.write.assert_called_once_with(cert_value.encode("ascii"))
        mock_named_temporary_file.return_value.flush.assert_called_once()

    @pytest.mark.parametrize("cert_name", ["sslcert", "sslkey", "sslrootcert"])
    @mock.patch(HOOK_STR.format("NamedTemporaryFile"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_set_temporary_ssl_file_exception(
        self,
        mock_get_connection,
        mock_named_temporary_file,
        cert_name,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        expected_file_name = "/test/path/to/file"
        mock_named_temporary_file.return_value.name = expected_file_name
        cert_value = "test-cert-value"
        source_cert_path = "/source/cert/path/to/file"

        hook = CloudSQLDatabaseHook()

        with pytest.raises(AirflowException):
            hook._set_temporary_ssl_file(
                cert_name=cert_name, cert_value=cert_value, cert_path=source_cert_path
            )

        assert not mock_named_temporary_file.called
        assert not mock_named_temporary_file.return_value.write.called
        assert not mock_named_temporary_file.return_value.flush.called

    @pytest.mark.parametrize("cert_name", ["sslcert", "sslkey", "sslrootcert"])
    @mock.patch(HOOK_STR.format("NamedTemporaryFile"))
    @mock.patch(HOOK_STR.format("CloudSQLDatabaseHook.get_connection"))
    def test_set_temporary_ssl_file_none(
        self,
        mock_get_connection,
        mock_named_temporary_file,
        cert_name,
    ):
        mock_get_connection.return_value = mock.MagicMock(
            extra_dejson={
                "database_type": "postgres",
                "location": "test",
                "instance": "instance",
            }
        )
        expected_file_name = "/test/path/to/file"
        mock_named_temporary_file.return_value.name = expected_file_name

        hook = CloudSQLDatabaseHook()

        actual_path = hook._set_temporary_ssl_file(cert_name=cert_name)

        assert actual_path is None
        assert not mock_named_temporary_file.called
        assert not mock_named_temporary_file.return_value.write.called
        assert not mock_named_temporary_file.return_value.flush.called


class TestCloudSqlDatabaseQueryHook:
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def setup_method(self, method, mock_get_conn):
        self.sql_connection = Connection(
            conn_id="my_gcp_sql_connection",
            conn_type="gcpcloudsql",
            login="login",
            password="password",
            host="host",
            schema="schema",
            extra='{"database_type":"postgres", "location":"my_location", '
            '"instance":"my_instance", "use_proxy": true, '
            '"project_id":"my_project"}',
        )
        self.connection = Connection(
            conn_id="my_gcp_connection",
            conn_type="google_cloud_platform",
        )
        scopes = [
            "https://www.googleapis.com/auth/pubsub",
            "https://www.googleapis.com/auth/datastore",
            "https://www.googleapis.com/auth/bigquery",
            "https://www.googleapis.com/auth/devstorage.read_write",
            "https://www.googleapis.com/auth/logging.write",
            "https://www.googleapis.com/auth/cloud-platform",
        ]
        conn_extra = {
            "scope": ",".join(scopes),
            "project": "your-gcp-project",
            "key_path": "/var/local/google_cloud_default.json",
        }
        conn_extra_json = json.dumps(conn_extra)
        if AIRFLOW_V_3_1_PLUS:
            self.connection.extra = conn_extra_json
        else:
            self.connection.set_extra(conn_extra_json)

        mock_get_conn.side_effect = [self.sql_connection, self.connection]
        self.db_hook = CloudSQLDatabaseHook(
            gcp_cloudsql_conn_id="my_gcp_sql_connection", gcp_conn_id="my_gcp_connection"
        )

    def test_get_sqlproxy_runner(self):
        self.db_hook._generate_connection_uri()
        sqlproxy_runner = self.db_hook.get_sqlproxy_runner()
        assert sqlproxy_runner.gcp_conn_id == self.connection.conn_id
        project = self.sql_connection.extra_dejson["project_id"]
        location = self.sql_connection.extra_dejson["location"]
        instance = self.sql_connection.extra_dejson["instance"]
        instance_spec = f"{project}:{location}:{instance}"
        assert sqlproxy_runner.instance_specification == instance_spec

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_not_too_long_unix_socket_path(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=postgres&"
            "project_id=example-project&location=europe-west1&"
            "instance="
            "test_db_with_longname_but_with_limit_of_UNIX_socket&"
            "use_proxy=True&sql_proxy_use_tcp=False"
        )
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "postgres"
        assert connection.schema == "testdb"

    def _verify_postgres_connection(self, get_connection, uri):
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "postgres"
        assert connection.host == "127.0.0.1"
        assert connection.port == 3200
        assert connection.schema == "testdb"
        return connection

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_postgres(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=postgres&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=False&use_ssl=False"
        )
        self._verify_postgres_connection(get_connection, uri)

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook._set_temporary_ssl_file")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_postgres_ssl(self, get_connection, mock_set_temporary_ssl_file):
        def side_effect_func(cert_name, cert_path, cert_value):
            return f"/tmp/{cert_name}"

        mock_set_temporary_ssl_file.side_effect = side_effect_func
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=postgres&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=False&use_ssl=True&sslcert=/bin/bash&"
            "sslkey=/bin/bash&sslrootcert=/bin/bash"
        )
        connection = self._verify_postgres_connection(get_connection, uri)
        assert connection.extra_dejson["sslkey"] == "/tmp/sslkey"
        assert connection.extra_dejson["sslcert"] == "/tmp/sslcert"
        assert connection.extra_dejson["sslrootcert"] == "/tmp/sslrootcert"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_postgres_proxy_socket(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=postgres&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=True&sql_proxy_use_tcp=False"
        )
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "postgres"
        assert tempfile.gettempdir() in connection.host
        assert "example-project:europe-west1:testdb" in connection.host
        assert connection.port is None
        assert connection.schema == "testdb"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_project_id_missing(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=mysql&"
            "location=europe-west1&instance=testdb&"
            "use_proxy=False&use_ssl=False"
        )
        self.verify_mysql_connection(get_connection, uri)

    def verify_mysql_connection(self, get_connection, uri):
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "mysql"
        assert connection.host == "127.0.0.1"
        assert connection.port == 3200
        assert connection.schema == "testdb"
        return connection

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_postgres_proxy_tcp(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=postgres&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=True&sql_proxy_use_tcp=True"
        )
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "postgres"
        assert connection.host == "127.0.0.1"
        assert connection.port != 3200
        assert connection.schema == "testdb"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_mysql(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=mysql&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=False&use_ssl=False"
        )
        self.verify_mysql_connection(get_connection, uri)

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook._set_temporary_ssl_file")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_mysql_ssl(self, get_connection, mock_set_temporary_ssl_file):
        def side_effect_func(cert_name, cert_path, cert_value):
            return f"/tmp/{cert_name}"

        mock_set_temporary_ssl_file.side_effect = side_effect_func
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=mysql&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=False&use_ssl=True&sslcert=/bin/bash&"
            "sslkey=/bin/bash&sslrootcert=/bin/bash"
        )
        connection = self.verify_mysql_connection(get_connection, uri)
        assert json.loads(connection.extra_dejson["ssl"])["cert"] == "/tmp/sslcert"
        assert json.loads(connection.extra_dejson["ssl"])["key"] == "/tmp/sslkey"
        assert json.loads(connection.extra_dejson["ssl"])["ca"] == "/tmp/sslrootcert"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_mysql_proxy_socket(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=mysql&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=True&sql_proxy_use_tcp=False"
        )
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "mysql"
        assert connection.host == "localhost"
        assert tempfile.gettempdir() in connection.extra_dejson["unix_socket"]
        assert "example-project:europe-west1:testdb" in connection.extra_dejson["unix_socket"]
        assert connection.port is None
        assert connection.schema == "testdb"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook.get_connection")
    def test_hook_with_correct_parameters_mysql_tcp(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=mysql&"
            "project_id=example-project&location=europe-west1&instance=testdb&"
            "use_proxy=True&sql_proxy_use_tcp=True"
        )
        if AIRFLOW_V_3_1_PLUS:
            get_connection.side_effect = [Connection(conn_id="test_conn_id", **_parse_from_uri(uri))]
        else:
            get_connection.side_effect = [Connection(uri=uri)]
        hook = CloudSQLDatabaseHook()
        connection = hook.create_connection()
        assert connection.conn_type == "mysql"
        assert connection.host == "127.0.0.1"
        assert connection.port != 3200
        assert connection.schema == "testdb"


def get_processor():
    processor = os.uname().machine
    if processor == "x86_64":
        processor = "amd64"
    if processor == "aarch64":
        processor = "arm64"
    return processor


class TestCloudSqlProxyRunner:
    @pytest.mark.parametrize(
        ("version", "download_url"),
        [
            (
                "v1.23.0",
                "https://storage.googleapis.com/cloudsql-proxy/v1.23.0/cloud_sql_proxy."
                f"{platform.system().lower()}.{get_processor()}",
            ),
            (
                "v1.23.0-preview.1",
                "https://storage.googleapis.com/cloudsql-proxy/v1.23.0-preview.1/cloud_sql_proxy."
                f"{platform.system().lower()}.{get_processor()}",
            ),
        ],
    )
    def test_cloud_sql_proxy_runner_version_ok(self, version, download_url):
        runner = CloudSqlProxyRunner(
            path_prefix="12345678",
            instance_specification="project:us-east-1:instance",
            sql_proxy_version=version,
        )
        assert runner._get_sql_proxy_download_url() == download_url

    @pytest.mark.parametrize(
        "version",
        [
            "v1.23.",
            "v1.23.0..",
            "v1.23.0\\",
            "\\",
        ],
    )
    def test_cloud_sql_proxy_runner_version_nok(self, version):
        runner = CloudSqlProxyRunner(
            path_prefix="12345678",
            instance_specification="project:us-east-1:instance",
            sql_proxy_version=version,
        )
        with pytest.raises(ValueError, match="The sql_proxy_version should match the regular expression"):
            runner._get_sql_proxy_download_url()


class TestCloudSQLAsyncHook:
    @pytest.mark.asyncio
    @mock.patch(HOOK_STR.format("CloudSQLAsyncHook._get_conn"))
    async def test_async_get_operation_name_should_execute_successfully(self, mocked_conn, hook_async):
        await hook_async.get_operation_name(
            operation_name=OPERATION_NAME,
            project_id=PROJECT_ID,
            session=session,
        )

        mocked_conn.assert_awaited_once_with(url=OPERATION_URL, session=session)

    @pytest.mark.asyncio
    @mock.patch(HOOK_STR.format("CloudSQLAsyncHook.get_operation_name"))
    async def test_async_get_operation_completed_should_execute_successfully(self, mocked_get, hook_async):
        response = aiohttp.ClientResponse(
            "get",
            URL(OPERATION_URL),
            request_info=mock.Mock(),
            writer=mock.Mock(),
            continue100=None,
            timer=TimerNoop(),
            traces=[],
            loop=mock.Mock(),
            session=None,
        )
        response.status = 200
        mocked_get.return_value = response
        mocked_get.return_value._headers = {"Authorization": "test-token"}
        mocked_get.return_value._body = b'{"status": "DONE"}'

        operation = await hook_async.get_operation(operation_name=OPERATION_NAME, project_id=PROJECT_ID)
        mocked_get.assert_awaited_once()
        assert operation["status"] == "DONE"

    @pytest.mark.asyncio
    @mock.patch(HOOK_STR.format("CloudSQLAsyncHook.get_operation_name"))
    async def test_async_get_operation_running_should_execute_successfully(self, mocked_get, hook_async):
        response = aiohttp.ClientResponse(
            "get",
            URL(OPERATION_URL),
            request_info=mock.Mock(),
            writer=mock.Mock(),
            continue100=None,
            timer=TimerNoop(),
            traces=[],
            loop=mock.Mock(),
            session=None,
        )
        response.status = 200
        mocked_get.return_value = response
        mocked_get.return_value._headers = {"Authorization": "test-token"}
        mocked_get.return_value._body = b'{"status": "RUNNING"}'

        operation = await hook_async.get_operation(operation_name=OPERATION_NAME, project_id=PROJECT_ID)
        mocked_get.assert_awaited_once()
        assert operation["status"] == "RUNNING"

    @pytest.mark.asyncio
    @mock.patch(HOOK_STR.format("CloudSQLAsyncHook._get_conn"))
    async def test_async_get_operation_exception_should_execute_successfully(
        self, mocked_get_conn, hook_async
    ):
        """Assets that the logging is done correctly when CloudSQLAsyncHook raises HttpError"""
        mocked_get_conn.side_effect = HttpError(
            resp=mock.MagicMock(status=409), content=b"Operation already exists"
        )
        with pytest.raises(HttpError):
            await hook_async.get_operation(operation_name=OPERATION_NAME, project_id=PROJECT_ID)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Cloud Composer sensor."""

from __future__ import annotations

import json
from collections.abc import Collection, Iterable, Sequence
from datetime import datetime, timedelta
from functools import cached_property
from typing import TYPE_CHECKING

from dateutil import parser
from google.api_core.exceptions import NotFound
from google.cloud.orchestration.airflow.service_v1.types import Environment, ExecuteAirflowCommandResponse

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.common.compat.sdk import AirflowSkipException, BaseSensorOperator
from airflow.providers.google.cloud.hooks.cloud_composer import CloudComposerHook
from airflow.providers.google.cloud.triggers.cloud_composer import (
    CloudComposerDAGRunTrigger,
    CloudComposerExternalTaskTrigger,
)
from airflow.providers.google.common.consts import GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME
from airflow.providers.standard.exceptions import (
    DuplicateStateError,
    ExternalDagFailedError,
    ExternalTaskFailedError,
    ExternalTaskGroupFailedError,
)
from airflow.utils.state import State, TaskInstanceState

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class CloudComposerDAGRunSensor(BaseSensorOperator):
    """
    Check if a DAG run has completed.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param environment_id: The name of the Composer environment.
    :param composer_dag_id: The ID of executable DAG.
    :param allowed_states: Iterable of allowed states, default is ``['success']``.
    :param execution_range: execution DAGs time range. Sensor checks DAGs states only for DAGs which were
        started in this time range. For yesterday, use [positive!] datetime.timedelta(days=1).
        For future, use [negative!] datetime.timedelta(days=-1). For specific time, use list of
        datetimes [datetime(2024,3,22,11,0,0), datetime(2024,3,22,12,0,0)].
        Or [datetime(2024,3,22,0,0,0)] in this case sensor will check for states from specific time in the
        past till current time execution.
        Default value datetime.timedelta(days=1).
    :param composer_dag_run_id: The Run ID of executable task. The 'execution_range' param is ignored, if both specified.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param poll_interval: Optional: Control the rate of the poll for the result of deferrable run.
    :param deferrable: Run sensor in deferrable mode.
    """

    template_fields = (
        "project_id",
        "region",
        "environment_id",
        "composer_dag_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        environment_id: str,
        composer_dag_id: str,
        allowed_states: Iterable[str] | None = None,
        execution_range: timedelta | list[datetime] | None = None,
        composer_dag_run_id: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        use_rest_api: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.environment_id = environment_id
        self.composer_dag_id = composer_dag_id
        self.allowed_states = list(allowed_states) if allowed_states else [TaskInstanceState.SUCCESS.value]
        self.execution_range = execution_range
        self.composer_dag_run_id = composer_dag_run_id
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval
        self.use_rest_api = use_rest_api

        if self.composer_dag_run_id and self.execution_range:
            self.log.warning(
                "The composer_dag_run_id parameter and execution_range parameter do not work together. This run will ignore execution_range parameter and count only specified composer_dag_run_id parameter."
            )

    def _get_logical_dates(self, context) -> tuple[datetime, datetime]:
        logical_date = context.get("logical_date", None)
        if logical_date is None:
            raise RuntimeError(
                "logical_date is None. Please make sure the sensor is not used in an asset-triggered Dag. "
                "CloudComposerDAGRunSensor was designed to be used in time-based scheduled Dags only, "
                "and asset-triggered Dags do not have logical_date. "
            )
        if isinstance(self.execution_range, timedelta):
            if self.execution_range < timedelta(0):
                return logical_date, logical_date - self.execution_range
            return logical_date - self.execution_range, logical_date
        if isinstance(self.execution_range, list) and len(self.execution_range) > 0:
            return self.execution_range[0], self.execution_range[1] if len(
                self.execution_range
            ) > 1 else logical_date
        return logical_date - timedelta(1), logical_date

    def poke(self, context: Context) -> bool:
        start_date, end_date = self._get_logical_dates(context)

        if datetime.now(end_date.tzinfo) < end_date:
            return False

        dag_runs = self._pull_dag_runs()

        if len(dag_runs) == 0:
            self.log.info("Dag runs are empty. Sensor waits for dag runs...")
            return False

        if self.composer_dag_run_id:
            self.log.info(
                "Sensor waits for allowed states %s for specified RunID: %s",
                self.allowed_states,
                self.composer_dag_run_id,
            )
            composer_dag_run_id_status = self._check_composer_dag_run_id_states(
                dag_runs=dag_runs,
            )
            return composer_dag_run_id_status
        self.log.info("Sensor waits for allowed states: %s", self.allowed_states)
        allowed_states_status = self._check_dag_runs_states(
            dag_runs=dag_runs,
            start_date=start_date,
            end_date=end_date,
        )

        return allowed_states_status

    def _pull_dag_runs(self) -> list[dict]:
        """Pull the list of dag runs."""
        if self.use_rest_api:
            try:
                environment = self.hook.get_environment(
                    project_id=self.project_id,
                    region=self.region,
                    environment_id=self.environment_id,
                    timeout=self.timeout,
                )
            except NotFound as not_found_err:
                self.log.info("The Composer environment %s does not exist.", self.environment_id)
                raise AirflowException(not_found_err)
            composer_airflow_uri = environment.config.airflow_uri

            self.log.info(
                "Pulling the DAG %s runs from the %s environment...",
                self.composer_dag_id,
                self.environment_id,
            )
            dag_runs_response = self.hook.get_dag_runs(
                composer_airflow_uri=composer_airflow_uri,
                composer_dag_id=self.composer_dag_id,
                timeout=self.timeout,
            )
            dag_runs = dag_runs_response["dag_runs"]
        else:
            cmd_parameters = (
                ["-d", self.composer_dag_id, "-o", "json"]
                if self._composer_airflow_version < 3
                else [self.composer_dag_id, "-o", "json"]
            )
            dag_runs_cmd = self.hook.execute_airflow_command(
                project_id=self.project_id,
                region=self.region,
                environment_id=self.environment_id,
                command="dags",
                subcommand="list-runs",
                parameters=cmd_parameters,
            )
            cmd_result = self.hook.wait_command_execution_result(
                project_id=self.project_id,
                region=self.region,
                environment_id=self.environment_id,
                execution_cmd_info=ExecuteAirflowCommandResponse.to_dict(dag_runs_cmd),
            )
            dag_runs = json.loads(cmd_result["output"][0]["content"])
        return dag_runs

    def _check_dag_runs_states(
        self,
        dag_runs: list[dict],
        start_date: datetime,
        end_date: datetime,
    ) -> bool:
        for dag_run in dag_runs:
            if (
                start_date.timestamp()
                < parser.parse(
                    dag_run["execution_date" if self._composer_airflow_version < 3 else "logical_date"]
                ).timestamp()
                < end_date.timestamp()
            ) and dag_run["state"] not in self.allowed_states:
                return False
        return True

    def _get_composer_airflow_version(self) -> int:
        """Return Composer Airflow version."""
        environment_obj = self.hook.get_environment(
            project_id=self.project_id,
            region=self.region,
            environment_id=self.environment_id,
        )
        environment_config = Environment.to_dict(environment_obj)
        image_version = environment_config["config"]["software_config"]["image_version"]
        return int(image_version.split("airflow-")[1].split(".")[0])

    def _check_composer_dag_run_id_states(self, dag_runs: list[dict]) -> bool:
        for dag_run in dag_runs:
            if (
                dag_run["dag_run_id" if self.use_rest_api else "run_id"] == self.composer_dag_run_id
                and dag_run["state"] in self.allowed_states
            ):
                return True
        return False

    def execute(self, context: Context) -> None:
        self._composer_airflow_version = self._get_composer_airflow_version()
        if self.deferrable:
            start_date, end_date = self._get_logical_dates(context)
            self.defer(
                timeout=timedelta(seconds=self.timeout) if self.timeout else None,
                trigger=CloudComposerDAGRunTrigger(
                    project_id=self.project_id,
                    region=self.region,
                    environment_id=self.environment_id,
                    composer_dag_id=self.composer_dag_id,
                    composer_dag_run_id=self.composer_dag_run_id,
                    start_date=start_date,
                    end_date=end_date,
                    allowed_states=self.allowed_states,
                    gcp_conn_id=self.gcp_conn_id,
                    impersonation_chain=self.impersonation_chain,
                    poll_interval=self.poll_interval,
                    composer_airflow_version=self._composer_airflow_version,
                    use_rest_api=self.use_rest_api,
                ),
                method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
            )
        super().execute(context)

    def execute_complete(self, context: Context, event: dict):
        if event and event["status"] == "error":
            raise AirflowException(event["message"])
        self.log.info("DAG %s has executed successfully.", self.composer_dag_id)

    @cached_property
    def hook(self) -> CloudComposerHook:
        return CloudComposerHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class CloudComposerExternalTaskSensor(BaseSensorOperator):
    """
    Waits for a different DAG, task group, or task to complete for a specific composer environment.

    If both `composer_external_task_group_id` and `composer_external_task_id` are ``None`` (default), the sensor
    waits for the DAG.
    Values for `composer_external_task_group_id` and `composer_external_task_id` can't be set at the same time.

    By default, the CloudComposerExternalTaskSensor will wait for the external task to
    succeed, at which point it will also succeed. However, by default it will
    *not* fail if the external task fails, but will continue to check the status
    until the sensor times out (thus giving you time to retry the external task
    without also having to clear the sensor).

    By default, the CloudComposerExternalTaskSensor will not skip if the external task skips.
    To change this, simply set ``skipped_states=[TaskInstanceState.SKIPPED]``.
    Note that if you are monitoring multiple tasks, and one enters error state
    and the other enters a skipped state, then the external task will react to
    whichever one it sees first. If both happen together, then the failed state
    takes priority.

    It is possible to alter the default behavior by setting states which
    cause the sensor to fail, e.g. by setting ``allowed_states=[DagRunState.FAILED]``
    and ``failed_states=[DagRunState.SUCCESS]`` you will flip the behaviour to
    get a sensor which goes green when the external task *fails* and immediately
    goes red if the external task *succeeds*!

    Note that ``soft_fail`` is respected when examining the failed_states. Thus
    if the external task enters a failed state and ``soft_fail == True`` the
    sensor will _skip_ rather than fail. As a result, setting ``soft_fail=True``
    and ``failed_states=[DagRunState.SKIPPED]`` will result in the sensor
    skipping if the external task skips. However, this is a contrived
    example---consider using ``skipped_states`` if you would like this
    behaviour. Using ``skipped_states`` allows the sensor to skip if the target
    fails, but still enter failed state on timeout. Using ``soft_fail == True``
    as above will cause the sensor to skip if the target fails, but also if it
    times out.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param environment_id: The name of the Composer environment.
    :param composer_external_dag_id: The dag_id that contains the task you want to
        wait for. (templated)
    :param composer_external_task_id: The task_id that contains the task you want to
        wait for. (templated)
    :param composer_external_task_ids: The list of task_ids that you want to wait for. (templated)
        If ``None`` (default value) the sensor waits for the DAG. Either
        composer_external_task_id or composer_external_task_ids can be passed to
        CloudComposerExternalTaskSensor, but not both.
    :param composer_external_task_group_id: The task_group_id that contains the task you want to
        wait for. (templated)
    :param allowed_states: Iterable of allowed states, default is ``['success']``
    :param skipped_states: Iterable of states to make this task mark as skipped, default is ``None``
    :param failed_states: Iterable of failed or dis-allowed states, default is ``None``
    :param execution_range: execution DAGs time range. Sensor checks DAGs states only for DAGs which were
        started in this time range. For yesterday, use [positive!] datetime.timedelta(days=1).
        For future, use [negative!] datetime.timedelta(days=-1). For specific time, use list of
        datetimes [datetime(2024,3,22,11,0,0), datetime(2024,3,22,12,0,0)].
        Or [datetime(2024,3,22,0,0,0)] in this case sensor will check for states from specific time in the
        past till current time execution.
        Default value datetime.timedelta(days=1).
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param poll_interval: Optional: Control the rate of the poll for the result of deferrable run.
    :param deferrable: Run sensor in deferrable mode.
    """

    template_fields = (
        "project_id",
        "region",
        "environment_id",
        "composer_external_dag_id",
        "composer_external_task_id",
        "composer_external_task_ids",
        "composer_external_task_group_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        environment_id: str,
        composer_external_dag_id: str,
        composer_external_task_id: str | None = None,
        composer_external_task_ids: Collection[str] | None = None,
        composer_external_task_group_id: str | None = None,
        allowed_states: Iterable[str] | None = None,
        skipped_states: Iterable[str] | None = None,
        failed_states: Iterable[str] | None = None,
        execution_range: timedelta | list[datetime] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.environment_id = environment_id

        self.allowed_states = list(allowed_states) if allowed_states else [TaskInstanceState.SUCCESS.value]
        self.skipped_states = list(skipped_states) if skipped_states else []
        self.failed_states = list(failed_states) if failed_states else []

        total_states = set(self.allowed_states + self.skipped_states + self.failed_states)

        if len(total_states) != len(self.allowed_states) + len(self.skipped_states) + len(self.failed_states):
            raise DuplicateStateError(
                "Duplicate values provided across allowed_states, skipped_states and failed_states."
            )

        # convert [] to None
        if not composer_external_task_ids:
            composer_external_task_ids = None

        # can't set both single task id and a list of task ids
        if composer_external_task_id is not None and composer_external_task_ids is not None:
            raise ValueError(
                "Only one of `composer_external_task_id` or `composer_external_task_ids` may "
                "be provided to CloudComposerExternalTaskSensor; "
                "use `composer_external_task_id` or `composer_external_task_ids` or `composer_external_task_group_id`."
            )

        # since both not set, convert the single id to a 1-elt list - from here on, we only consider the list
        if composer_external_task_id is not None:
            composer_external_task_ids = [composer_external_task_id]

        if composer_external_task_group_id is not None and composer_external_task_ids is not None:
            raise ValueError(
                "Only one of `composer_external_task_group_id` or `composer_external_task_ids` may "
                "be provided to CloudComposerExternalTaskSensor; "
                "use `composer_external_task_id` or `composer_external_task_ids` or `composer_external_task_group_id`."
            )

        # check the requested states are all valid states for the target type, be it dag or task
        if composer_external_task_ids or composer_external_task_group_id:
            if not total_states <= set(State.task_states):
                raise ValueError(
                    "Valid values for `allowed_states`, `skipped_states` and `failed_states` "
                    "when `composer_external_task_id` or `composer_external_task_ids` or `composer_external_task_group_id` "
                    f"is not `None`: {State.task_states}"
                )
        elif not total_states <= set(State.dag_states):
            raise ValueError(
                "Valid values for `allowed_states`, `skipped_states` and `failed_states` "
                f"when `composer_external_task_id` and `composer_external_task_group_id` is `None`: {State.dag_states}"
            )

        self.execution_range = execution_range
        self.composer_external_dag_id = composer_external_dag_id
        self.composer_external_task_id = composer_external_task_id
        self.composer_external_task_ids = composer_external_task_ids
        self.composer_external_task_group_id = composer_external_task_group_id
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    def _get_logical_dates(self, context) -> tuple[datetime, datetime]:
        logical_date = context.get("logical_date", None)
        if logical_date is None:
            raise RuntimeError(
                "logical_date is None. Please make sure the sensor is not used in an asset-triggered Dag. "
                "CloudComposerDAGRunSensor was designed to be used in time-based scheduled Dags only, "
                "and asset-triggered Dags do not have logical_date. "
            )
        if isinstance(self.execution_range, timedelta):
            if self.execution_range < timedelta(0):
                return logical_date, logical_date - self.execution_range
            return logical_date - self.execution_range, logical_date
        if isinstance(self.execution_range, list) and len(self.execution_range) > 0:
            return self.execution_range[0], self.execution_range[1] if len(
                self.execution_range
            ) > 1 else logical_date
        return logical_date - timedelta(1), logical_date

    def poke(self, context: Context) -> bool:
        start_date, end_date = self._get_logical_dates(context)

        task_instances = self._get_task_instances(
            start_date=start_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
            end_date=end_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        )

        if len(task_instances) == 0:
            self.log.info("Task Instances are empty. Sensor waits for task instances...")
            return False

        if self.failed_states:
            external_task_status = self._check_task_instances_states(
                task_instances=task_instances,
                start_date=start_date,
                end_date=end_date,
                states=self.failed_states,
            )
            self._handle_failed_states(external_task_status)

        if self.skipped_states:
            external_task_status = self._check_task_instances_states(
                task_instances=task_instances,
                start_date=start_date,
                end_date=end_date,
                states=self.skipped_states,
            )
            self._handle_skipped_states(external_task_status)

        self.log.info("Sensor waits for allowed states: %s", self.allowed_states)
        external_task_status = self._check_task_instances_states(
            task_instances=task_instances,
            start_date=start_date,
            end_date=end_date,
            states=self.allowed_states,
        )
        return external_task_status

    def _get_task_instances(self, start_date: str, end_date: str) -> list[dict]:
        """Get the list of task instances."""
        try:
            environment = self.hook.get_environment(
                project_id=self.project_id,
                region=self.region,
                environment_id=self.environment_id,
                timeout=self.timeout,
            )
        except NotFound as not_found_err:
            self.log.info("The Composer environment %s does not exist.", self.environment_id)
            raise AirflowException(not_found_err)
        composer_airflow_uri = environment.config.airflow_uri

        self.log.info(
            "Pulling the DAG '%s' task instances from the '%s' environment...",
            self.composer_external_dag_id,
            self.environment_id,
        )
        task_instances_response = self.hook.get_task_instances(
            composer_airflow_uri=composer_airflow_uri,
            composer_dag_id=self.composer_external_dag_id,
            query_parameters={
                "execution_date_gte"
                if self._composer_airflow_version < 3
                else "logical_date_gte": start_date,
                "execution_date_lte" if self._composer_airflow_version < 3 else "logical_date_lte": end_date,
            },
            timeout=self.timeout,
        )
        task_instances = task_instances_response["task_instances"]

        if self.composer_external_task_ids:
            task_instances = [
                task_instance
                for task_instance in task_instances
                if task_instance["task_id"] in self.composer_external_task_ids
            ]
        elif self.composer_external_task_group_id:
            task_instances = [
                task_instance
                for task_instance in task_instances
                if self.composer_external_task_group_id in task_instance["task_id"].split(".")
            ]

        return task_instances

    def _check_task_instances_states(
        self,
        task_instances: list[dict],
        start_date: datetime,
        end_date: datetime,
        states: Iterable[str],
    ) -> bool:
        for task_instance in task_instances:
            if (
                start_date.timestamp()
                < parser.parse(
                    task_instance["execution_date" if self._composer_airflow_version < 3 else "logical_date"]
                ).timestamp()
                < end_date.timestamp()
            ) and task_instance["state"] not in states:
                return False
        return True

    def _get_composer_airflow_version(self) -> int:
        """Return Composer Airflow version."""
        environment_obj = self.hook.get_environment(
            project_id=self.project_id,
            region=self.region,
            environment_id=self.environment_id,
        )
        environment_config = Environment.to_dict(environment_obj)
        image_version = environment_config["config"]["software_config"]["image_version"]
        return int(image_version.split("airflow-")[1].split(".")[0])

    def _handle_failed_states(self, failed_status: bool) -> None:
        """Handle failed states and raise appropriate exceptions."""
        if failed_status:
            if self.composer_external_task_ids:
                if self.soft_fail:
                    raise AirflowSkipException(
                        f"Some of the external tasks '{self.composer_external_task_ids}' "
                        f"in DAG '{self.composer_external_dag_id}' failed. Skipping due to soft_fail."
                    )
                raise ExternalTaskFailedError(
                    f"Some of the external tasks '{self.composer_external_task_ids}' "
                    f"in DAG '{self.composer_external_dag_id}' failed."
                )
            if self.composer_external_task_group_id:
                if self.soft_fail:
                    raise AirflowSkipException(
                        f"The external task_group '{self.composer_external_task_group_id}' "
                        f"in DAG '{self.composer_external_dag_id}' failed. Skipping due to soft_fail."
                    )
                raise ExternalTaskGroupFailedError(
                    f"The external task_group '{self.composer_external_task_group_id}' "
                    f"in DAG '{self.composer_external_dag_id}' failed."
                )
            if self.soft_fail:
                raise AirflowSkipException(
                    f"The external DAG '{self.composer_external_dag_id}' failed. Skipping due to soft_fail."
                )
            raise ExternalDagFailedError(f"The external DAG '{self.composer_external_dag_id}' failed.")

    def _handle_skipped_states(self, skipped_status: bool) -> None:
        """Handle skipped states and raise appropriate exceptions."""
        if skipped_status:
            if self.composer_external_task_ids:
                raise AirflowSkipException(
                    f"Some of the external tasks '{self.composer_external_task_ids}' "
                    f"in DAG '{self.composer_external_dag_id}' reached a state in our states-to-skip-on list. Skipping."
                )
            if self.composer_external_task_group_id:
                raise AirflowSkipException(
                    f"The external task_group '{self.composer_external_task_group_id}' "
                    f"in DAG '{self.composer_external_dag_id}' reached a state in our states-to-skip-on list. Skipping."
                )
            raise AirflowSkipException(
                f"The external DAG '{self.composer_external_dag_id}' reached a state in our states-to-skip-on list. "
                "Skipping."
            )

    def execute(self, context: Context) -> None:
        self._composer_airflow_version = self._get_composer_airflow_version()

        if self.composer_external_task_ids and len(self.composer_external_task_ids) > len(
            set(self.composer_external_task_ids)
        ):
            raise ValueError("Duplicate task_ids passed in composer_external_task_ids parameter")

        if self.composer_external_task_ids:
            self.log.info(
                "Poking for tasks '%s' in dag '%s' on Composer environment '%s' ... ",
                self.composer_external_task_ids,
                self.composer_external_dag_id,
                self.environment_id,
            )

        if self.composer_external_task_group_id:
            self.log.info(
                "Poking for task_group '%s' in dag '%s' on Composer environment '%s' ... ",
                self.composer_external_task_group_id,
                self.composer_external_dag_id,
                self.environment_id,
            )

        if (
            self.composer_external_dag_id
            and not self.composer_external_task_group_id
            and not self.composer_external_task_ids
        ):
            self.log.info(
                "Poking for DAG '%s' on Composer environment '%s' ... ",
                self.composer_external_dag_id,
                self.environment_id,
            )

        if self.deferrable:
            start_date, end_date = self._get_logical_dates(context)
            self.defer(
                timeout=timedelta(seconds=self.timeout) if self.timeout else None,
                trigger=CloudComposerExternalTaskTrigger(
                    project_id=self.project_id,
                    region=self.region,
                    environment_id=self.environment_id,
                    composer_external_dag_id=self.composer_external_dag_id,
                    composer_external_task_ids=self.composer_external_task_ids,
                    composer_external_task_group_id=self.composer_external_task_group_id,
                    start_date=start_date,
                    end_date=end_date,
                    allowed_states=self.allowed_states,
                    skipped_states=self.skipped_states,
                    failed_states=self.failed_states,
                    gcp_conn_id=self.gcp_conn_id,
                    impersonation_chain=self.impersonation_chain,
                    poll_interval=self.poll_interval,
                    composer_airflow_version=self._composer_airflow_version,
                ),
                method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
            )
        super().execute(context)

    def execute_complete(self, context: Context, event: dict):
        if event and event["status"] == "error":
            raise AirflowException(event["message"])
        if event and event["status"] == "failed":
            self._handle_failed_states(True)
        elif event and event["status"] == "skipped":
            self._handle_skipped_states(True)

        self.log.info("External tasks for DAG '%s' has executed successfully.", self.composer_external_dag_id)

    @cached_property
    def hook(self) -> CloudComposerHook:
        return CloudComposerHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
